/*----------------------------------------------------------*
   DXMIO Dynamixel スレーブ
                                       Last Edit '17 11/16
   Copyright (c) BestTechnology CO.,LTD. 2017
 *----------------------------------------------------------*/

#define DX_PROTOCOL_VER 2 // DXプロトコルバージョン選択 (1 or 2)

#include  <chip.h>
#include  <us.h>
#if DX_PROTOCOL_VER == 1
#include  <dxclient.h>
#elif DX_PROTOCOL_VER == 2
#include  <dx2client.h>
#endif
#include  <piocfg.h>
#include  <wl_delay.h>

#include  <FreeRTOS.h>
#include  <task.h>

/*======================================================================*/
/* マクロ・グローバル変数など                                           */
/*======================================================================*/
//#define _USE_MYTASK

// 基本機能端子割り当て
#define _TXD  (16)
#define _RXD  (26)
#define _RTS  (27)
#define _LED  (24)

static const TPin pins[] = {
  { _TXD,  PIO_TYPE_MOVABLE,  SWM_U0_TXD_O, PIO_MODE_DEFAULT },
  { _RXD,  PIO_TYPE_MOVABLE,  SWM_U0_RXD_I, PIO_MODE_PULLUP  },
  { _RTS,  PIO_TYPE_MOVABLE,  SWM_U0_RTS_O, PIO_MODE_DEFAULT },
  { _LED,  PIO_TYPE_OUTPUT_1, 0,            PIO_MODE_DEFAULT },
};

// Pin Mode
typedef enum { tPinDIN = 0, tPinDIN_PU, tPinDIN_PD, tPinDOUT, tPinDOUT_OC, tPinAIN, tPinPWM0, tPinPWM1, tPinPWM2, tPinPWM3, tPinPWM4, tPinPWM5, tPinMPW0, tPinMPW1, tPinMPW2, tPinMPW3, tPinMPC0, tPinMPC1, tPinMPC2, tPinMPC3, tPinSPFUNC } TMIOPinMode;

// Pin information
typedef struct {
  uint8_t               No;
  CHIP_PINx_T           CPin;
  CHIP_SWM_PIN_FIXED_T  AD;
} TPinInfo;

// 端子機能
static const TPinInfo PinInfo[12] = {
  {  7, IOCON_PIO7,  SWM_FIXED_ADC0  },
  {  6, IOCON_PIO6,  SWM_FIXED_ADC1  },
  { 14, IOCON_PIO14, SWM_FIXED_ADC2  },
  { 23, IOCON_PIO23, SWM_FIXED_ADC3  },
  { 22, IOCON_PIO22, SWM_FIXED_ADC4  },
  { 21, IOCON_PIO21, SWM_FIXED_ADC5  },
  { 20, IOCON_PIO20, SWM_FIXED_ADC6  },
  { 19, IOCON_PIO19, SWM_FIXED_ADC7  },
  { 18, IOCON_PIO18, SWM_FIXED_ADC8  },
  { 17, IOCON_PIO17, SWM_FIXED_ADC9  },
  { 13, IOCON_PIO13, SWM_FIXED_ADC10 },
  {  4, IOCON_PIO4,  SWM_FIXED_ADC11 }
};

static uint8_t previous_setting[12] = { tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN };

// シリアル通信バッファ
static uint8_t txd[50], rxd[150];

//-----------------------------
// コントロールテーブル
//-----------------------------
typedef struct {
  uint16_t  ModelNumber;
#if DX_PROTOCOL_VER == 2
  uint32_t  ModelInfomation;
#endif
  uint8_t   FirmwareVersion;
  uint8_t   ID;
  uint8_t   Baudrate;
  uint8_t   WriteNVM;

  uint8_t   LED;
  uint8_t   PinCfg[12];
  uint8_t   reserve1;
  uint16_t  PWMFrequency;
  uint16_t  PWMDuty[6];
  uint8_t   reserve2[10];
  uint16_t  Capture[4];
  uint16_t  OUT;
  uint16_t  IN;
  uint16_t  ADV[12];
  uint8_t   userarea[16];
} __attribute__((packed)) TE097CtrlTable;

TE097CtrlTable  *pmemmap;

#define _ADDR_PinCfg        ((int)&((TE097CtrlTable*)0)->PinCfg)
#define _ADDR_PWMFrequency  ((int)&((TE097CtrlTable*)0)->PWMFrequency)
#define _ADDR_PWMDuty       ((int)&((TE097CtrlTable*)0)->PWMDuty)
#define _ADDR_Capture       ((int)&((TE097CtrlTable*)0)->Capture)
#define _ADDR_OUT           ((int)&((TE097CtrlTable*)0)->OUT)
#define _ADDR_IN            ((int)&((TE097CtrlTable*)0)->IN)
#define _ADDR_ADV           ((int)&((TE097CtrlTable*)0)->ADV)
#define _ADDR_userarea      ((int)&((TE097CtrlTable*)0)->userarea)

//-----------------------------
// コントロールテーブルデフォルト値
//-----------------------------
static const TE097CtrlTable DEFAULT_DATA = {
  ModelNumber     :0x4020,
#if DX_PROTOCOL_VER == 2
  ModelInfomation :0,
#endif
  FirmwareVersion :'B',
  ID              :200,
#if DX_PROTOCOL_VER == 1
  Baudrate        :1,
#elif DX_PROTOCOL_VER == 2
  Baudrate        :3,
#endif
  WriteNVM        :0,
  LED             :0,
};

/*======================================================================*/
/* ETC                                                                  */
/*======================================================================*/
//-----------------------------
// LED
//-----------------------------
static void ToggleLED (void) {
  Chip_GPIO_SetPinToggle (LPC_GPIO_PORT, 0, _LED);
}

static void SetLED (bool On) {
  Chip_GPIO_SetPinState (LPC_GPIO_PORT, 0, _LED, (bool)!On);
}

//-----------------------------
// PWM
//-----------------------------
static void init_SCT (void) {
  static bool inited = false;

  if (!inited) {
    Chip_SCT_Init (LPC_SCT);
    inited = true;
  }
}

static void update_PWMFreqency (uint16_t freq) {
  static uint16_t oldfreq = 0;

  if (freq == 0) {
    Chip_SCTPWM_Stop (LPC_SCT);
    Chip_SCT_DeInit (LPC_SCT);
  } else {
    init_SCT ();
    if (oldfreq != freq) Chip_SCTPWM_SetRate (LPC_SCT, freq);
  }
  oldfreq = freq;
}

static void update_PWMDuty (uint8_t ch, uint16_t duty) {
  uint32_t d;

  Chip_SCTPWM_SetOutPin (LPC_SCT, ch + 1, ch);
  if (duty == 0)          d = 0;
  else if (duty == 65535) d = Chip_SCTPWM_GetTicksPerCycle (LPC_SCT) + 1;
  else                    d = ((uint64_t)duty * (uint64_t)Chip_SCTPWM_GetTicksPerCycle (LPC_SCT)) / (uint64_t)(65535UL);
  Chip_SCTPWM_SetDutyCycle (LPC_SCT, ch + 1, d);
  Chip_SCTPWM_Start (LPC_SCT);
}

//-----------------------------
// Capture
//-----------------------------
static void init_SCTCapture (void) {
  init_SCT ();

#if 0
  /* Initialize the SCT clock and reset the SCT */
  Chip_Clock_EnablePeriphClock(SYSCTL_CLOCK_SCT);
  Chip_SYSCTL_PeriphReset(RESET_SCT);
  Chip_INMUX_SetSCTInMux(LPC_INMUX, SCT_INMUX_0, SCT_INP_IN0);

  /* Enable flag to request an interrupt for Event 0 */
  Chip_SCT_EnableEventInt(LPC_SCT, SCT_EVT_0);

  /* Enable the interrupt for the SCT */
  NVIC_EnableIRQ(SCT_IRQn);

  /* Start the SCT counter by clearing Halt_L in the SCT control register */
  Chip_SCT_ClearControl(LPC_SCT, SCT_CTRL_HALT_L);


  LPC_SCT->CONFIG |= 0x1;                            // unified timer

  LPC_SCT->REGMODE_U = 0x1;														//match0 operates as capture register
  LPC_SCT->CAPCTRL[0].U = 0x1;													//event 0 causes capture 0	                                              

  LPC_SCT->EV[0].STATE = 0x00000001;                  // ev 0 happens in state 0
  LPC_SCT->EV[0].CTRL  = (1 << 0)  |                  // related to match 0
                             (0 << 5) |							//select input
                             (0 << 6) |							//select sct_input0 
                             (2 << 10) |							//falling edge																
                             (2 << 12) |              // input condition only
                             (1 << 14) |              // STATEV is new state
                             (0 << 15);               // STATEV[15] = 0
#else
//#define PWM_FREQUENCY 10000 // PWM frequency in Hz
//#define PWM_RESOLUTION_NS 1000 // Timer resolution in ns
#define PWM_FREQUENCY 50        // PWM frequency in Hz
#define PWM_RESOLUTION_NS 100  // Timer resolution in ns
#define PWM_MIN_DUTY_PERCENT 5  // Minimum allowed duty cycle in %
#define PWM_MAX_DUTY_PERCENT 80 // Maximum allowed duty cycle in %

/* Derived constants */
#define SCT_PRESCALER           (((SystemCoreClock / 1000u) * PWM_RESOLUTION_NS) / 1000000u - 1u)

#define match_min_width         ((10000000u * PWM_MIN_DUTY_PERCENT) / (PWM_FREQUENCY * PWM_RESOLUTION_NS))
#define match_max_width         ((10000000u * PWM_MAX_DUTY_PERCENT) / (PWM_FREQUENCY * PWM_RESOLUTION_NS))
#define match_no_input          ((10000000u * 300                 ) / (PWM_FREQUENCY * PWM_RESOLUTION_NS))

	Chip_INMUX_SetSCTInMux(LPC_INMUX, SCT_INMUX_0, SCT_INP_IN0);

  LPC_SCT->CONFIG |= (1 << 0) | (1 << 17); // unified, auto limit
  LPC_SCT->CTRL_U |= (SCT_PRESCALER << 5); // set prescaler
  LPC_SCT->REGMODE_L = 0x00000018; // 3x MATCH, 2x CAPTURE used
  LPC_SCT->MATCH[0].U = match_max_width; // match_max_width
  LPC_SCT->MATCHREL[0].U = match_max_width;
  LPC_SCT->MATCH[1].U = match_min_width; // match_min_width
  LPC_SCT->MATCHREL[1].U = match_min_width;
  LPC_SCT->MATCH[2].U = match_no_input; // match_no_input
  LPC_SCT->MATCHREL[2].U = match_no_input;
  LPC_SCT->EV[0].STATE = 0xFFFFFFFF; // event 0 happens in all states
  LPC_SCT->EV[0].CTRL = (2 << 0) | (1 << 12); // related to match_no_input only
  LPC_SCT->EV[1].STATE = 0xFFFFFFFF; // event 1 happens in all states
  LPC_SCT->EV[1].CTRL = (1 << 10) | (2 << 12); // IN_0 rising edge condition only
  LPC_SCT->EV[2].STATE = (1 << 0); // event 2 happens in state 0
  LPC_SCT->EV[2].CTRL = (1 << 0) | // related to match_min_width
  (1 << 12) | // match condition only
  (1 << 14) | // STATEV is loaded into state
  (1 << 15); // new state is 1
  LPC_SCT->EV[3].STATE = (1 << 1); // event 3 happens in state 1
  LPC_SCT->EV[3].CTRL = (2 << 10) | // IN_0 falling edge
  (2 << 12) | // IO condition only
  (1 << 14) | // STATEV is loaded into state
  (0 << 15); // new state is 0
  LPC_SCT->EV[4].STATE = (1 << 1); // event 4 happens in state 1
  LPC_SCT->EV[4].CTRL = (0 << 0) | // related to match_max_width
  (1 << 12) | // match condition only
  (1 << 14) | // STATEV is loaded into state
  (0 << 15); // new state is 0
  LPC_SCT->EV[5].STATE = (1 << 0); // event 5 happens in state 0
  LPC_SCT->EV[5].CTRL = (2 << 10) | (2 << 12); // IN_0 falling edge condition only
  LPC_SCT->CAPCTRL[3].U = (1 << 1); // event 1 is causing capture 3
  LPC_SCT->CAPCTRL[4].U = (1 << 3) | (1 << 5); // event 3 and 5 cause capture 4
  LPC_SCT->OUT[0].SET = (1 << 1); // event 1 set OUT0 (no timeout)
  LPC_SCT->OUT[0].CLR = (1 << 0); // event 0 clear OUT0 (timeout)
  LPC_SCT->OUT[1].SET = (1 << 3); // event 3 set OUT1 (no width error)
  LPC_SCT->OUT[1].CLR = (1 << 0) | (1 << 5); // event 0 and 5 clear OUT1 (width error)
  LPC_SCT->OUTPUT |= 3; // default set OUT0 and OUT1
  LPC_SCT->LIMIT_L = (1 << 0) | (1 << 1); // event 0 and 1 limit the timer
  LPC_SCT->EVEN = (1 << 0) | (1 << 5); // event 0 and 5 generate an irq
  NVIC_EnableIRQ(SCT_IRQn); // enable SCTimer/PWM interrupt
  LPC_SCT->CTRL_U &= ~(1 << 2); // start timer
#endif
}

uint16_t pwd[4];

void SCT_IRQHandler(void) {
#if 0
//	cap_flag = 1;
  volatile static uint32_t oldcap = 0;
  uint32_t cap = LPC_SCT->CAP[0].U, n;

//  if (cap >= oldcap) {
    n = cap - oldcap;
    pwd[0] = n;
    pwd[1] = n >> 16;
//  } else {
//    n = (UINT32_MAX - oldcap) + cap;
//    pwd[0] = n;
//    pwd[1] = n >> 16;
//  }
  oldcap = cap;
	/* Clear the SCT Event 0 Interrupt */
	Chip_SCT_ClearEventFlag(LPC_SCT, SCT_EVT_0);
#else
#define SCT_EVT_5                 (1 << 5)

	uint32_t status = LPC_SCT->EVEN;

	if (status & SCT_EVT_0)								// EVENT 0: timeout
	{
			/* New measurement result */
      pwd[2]=0x0400;
			;
	}
	if (status & SCT_EVT_5)								// EVENT 5: width error
	{
			/* Time-out (no signal) */
			;
      pwd[2]=0x8000;
	}

	LPC_SCT->EVFLAG = status;							// Acknowledge interrupts
#endif
}

//-----------------------------
// pin configuration
//-----------------------------
static void pin_config (uint8_t adch, TMIOPinMode pm, uint16_t freq, uint16_t duty) {
  static const uint8_t SCT_OUT_pin_movable[6] = { SWM_SCT_OUT0_O, SWM_SCT_OUT1_O, SWM_SCT_OUT2_O, SWM_SCT_OUT3_O, SWM_SCT_OUT4_O, SWM_SCT_OUT5_O };
  static const uint8_t SCT_IN_pin_movable[4] = { SWM_SCT_IN0_I, SWM_SCT_IN1_I, SWM_SCT_IN2_I, SWM_SCT_IN3_I };
  TPin pin = { PinInfo[adch].No, PIO_TYPE_FIXED, 0, PIO_MODE_DEFAULT };
  uint8_t ch;

  // 割り当て済みSCTの切断
  if ((previous_setting[adch] >= tPinPWM0) && (previous_setting[adch] <= tPinPWM5)) {
    ch = previous_setting[adch] - tPinPWM0;
    Chip_Clock_EnablePeriphClock (SYSCTL_CLOCK_SWM);
    Chip_SWM_MovablePinAssign (SCT_OUT_pin_movable[ch], 0xff);
    Chip_Clock_DisablePeriphClock (SYSCTL_CLOCK_SWM);
  }
  if ((previous_setting[adch] >= tPinMPW0) && (previous_setting[adch] <= tPinMPW3)) {
    ch = previous_setting[adch] - tPinMPW0;
    Chip_Clock_EnablePeriphClock (SYSCTL_CLOCK_SWM);
    Chip_SWM_MovablePinAssign (SCT_IN_pin_movable[ch], 0xff);
    Chip_Clock_DisablePeriphClock (SYSCTL_CLOCK_SWM);
  }
  if ((previous_setting[adch] >= tPinMPC0) && (previous_setting[adch] <= tPinMPC3)) {
    ch = previous_setting[adch] - tPinMPC0;
    Chip_Clock_EnablePeriphClock (SYSCTL_CLOCK_SWM);
    Chip_SWM_MovablePinAssign (SCT_IN_pin_movable[ch], 0xff);
    Chip_Clock_DisablePeriphClock (SYSCTL_CLOCK_SWM);
  }

  pin.gpio_no = PinInfo[adch].No;
  switch(pm) {
    case tPinDIN:     // 0
      PIO_Configure (&pin, 1);
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinDIN_PU:  // 1
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_PULLUP;
      break;
    case tPinDIN_PD:  // 2
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_PULLDOWN;
      break;
    case tPinDOUT:    // 3
      pin.pin_type = PIO_TYPE_OUTPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinDOUT_OC: // 4
      pin.pin_type = PIO_TYPE_OUTPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinAIN:     // 5
      pin.pin_type = PIO_TYPE_FIXED;
      pin.pin_movable_fixed = PinInfo[adch].AD;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinPWM0:    // 6
    case tPinPWM1:    // 7
    case tPinPWM2:    // 8
    case tPinPWM3:    // 9
    case tPinPWM4:    // 10
    case tPinPWM5:    // 11
      ch = pm - tPinPWM0;
      // disconnect
      Chip_Clock_EnablePeriphClock (SYSCTL_CLOCK_SWM);
      Chip_SWM_MovablePinAssign (SCT_OUT_pin_movable[ch], 0xff);
      Chip_Clock_DisablePeriphClock (SYSCTL_CLOCK_SWM);

      pin.pin_type = PIO_TYPE_MOVABLE;
      pin.pin_movable_fixed = SCT_OUT_pin_movable[ch];
      pin.pin_mode = PIO_MODE_DEFAULT;
      update_PWMFreqency (freq);
      update_PWMDuty (ch, duty);
      break;
    case tPinMPW0:    // 12
    case tPinMPW1:    // 13
    case tPinMPW2:    // 14
    case tPinMPW3:    // 15
      ch = pm - tPinMPW0;
      // disconnect
      Chip_Clock_EnablePeriphClock (SYSCTL_CLOCK_SWM);
      Chip_SWM_MovablePinAssign (SCT_IN_pin_movable[ch], 0xff);
      Chip_Clock_DisablePeriphClock (SYSCTL_CLOCK_SWM);

      pin.pin_type = PIO_TYPE_MOVABLE;
      pin.pin_movable_fixed = SCT_IN_pin_movable[ch];
      pin.pin_mode = PIO_MODE_DEFAULT;
//      init_SCTCapture ();
      break;
    case tPinMPC0:    // 16
    case tPinMPC1:    // 17
    case tPinMPC2:    // 18
    case tPinMPC3:    // 19
      ch = pm - tPinMPC0;
      // disconnect
      Chip_Clock_EnablePeriphClock (SYSCTL_CLOCK_SWM);
      Chip_SWM_MovablePinAssign (SCT_IN_pin_movable[ch], 0xff);
      Chip_Clock_DisablePeriphClock (SYSCTL_CLOCK_SWM);

      pin.pin_type = PIO_TYPE_MOVABLE;
      pin.pin_movable_fixed = SCT_IN_pin_movable[ch];
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinSPFUNC:  // 20
      break;
  }
  previous_setting[adch] = pm;
  PIO_Configure (&pin, 1);
}

//-----------------------------
// Pin config
//-----------------------------
static bool SetPinAssign (CHIP_SWM_PIN_MOVABLE_T movable, int pin) {
  if ((pin < 12) && (movable != SWM_U0_TXD_O) && (movable != SWM_U0_RXD_I) && (movable != SWM_U0_RTS_O)) {
    Chip_Clock_EnablePeriphClock (SYSCTL_CLOCK_SWM);
    Chip_SWM_DisableFixedPin (PinInfo[pin].AD);
    Chip_SWM_MovablePinAssign (movable, PinInfo[pin].No);
    Chip_Clock_DisablePeriphClock (SYSCTL_CLOCK_SWM);

    return true;
  }

  return false;
}

//-----------------------------
// WDT
//-----------------------------
static void init_wdt (void) {
	uint32_t wdtFreq;

  Chip_Clock_SetWDTOSC (WDTLFO_OSC_0_60, 64);      // Freq = 0.6Mhz, divided by 64. WDT_OSC should be 9.375khz
  Chip_SYSCTL_PowerUp (SYSCTL_SLPWAKE_WDTOSC_PD);  // Enable the power to the WDT
  wdtFreq = Chip_Clock_GetWDTOSCRate () / 4;       // The WDT divides the input frequency into it by 4
  Chip_WWDT_Init (LPC_WWDT);                       // Initialize WWDT (also enables WWDT clock)
  /* Set watchdog feed time constant to approximately 2s
     Set watchdog warning time to 512 ticks after feed time constant
     Set watchdog window time to 3s */
  Chip_WWDT_SetTimeOut (LPC_WWDT, wdtFreq * 2);
  Chip_WWDT_SetWarning (LPC_WWDT, 512);
  Chip_WWDT_SetWindow (LPC_WWDT, wdtFreq * 3);
  Chip_WWDT_SetOption (LPC_WWDT, WWDT_WDMOD_WDRESET);  // Configure WWDT to reset on timeout
  Chip_WWDT_ClearStatusFlag (LPC_WWDT, WWDT_WDMOD_WDTOF | WWDT_WDMOD_WDINT); // Clear watchdog warning and timeout interrupts
  Chip_WWDT_Start (LPC_WWDT);  // Start watchdog
}

//-----------------------------
// ADC
//-----------------------------
static void init_adc(void) {
  // Setup ADC for 12-bit mode and normal power
  Chip_ADC_Init (LPC_ADC, 0);
  // Need to do a calibration after initialization and trim
  Chip_ADC_StartCalibration (LPC_ADC);
  while (!Chip_ADC_IsCalibrationDone (LPC_ADC)) ;
  // Setup for maximum ADC clock rate using sycnchronous clocking
  Chip_ADC_SetClockRate (LPC_ADC, 10000);  // 10kHz
  // Setup a sequencer to do the following
  Chip_ADC_SetupSequencer (LPC_ADC, ADC_SEQA_IDX, (ADC_SEQ_CTRL_CHANSEL(0) | ADC_SEQ_CTRL_CHANSEL(1) | ADC_SEQ_CTRL_CHANSEL(2) | ADC_SEQ_CTRL_CHANSEL(3) | ADC_SEQ_CTRL_CHANSEL(4) | ADC_SEQ_CTRL_CHANSEL(5) | ADC_SEQ_CTRL_CHANSEL(6) | ADC_SEQ_CTRL_CHANSEL(7) | ADC_SEQ_CTRL_CHANSEL(8) | ADC_SEQ_CTRL_CHANSEL(9) | ADC_SEQ_CTRL_CHANSEL(10) | ADC_SEQ_CTRL_CHANSEL(11) | ADC_SEQ_CTRL_BURST));
  Chip_ADC_StartBurstSequencer (LPC_ADC, ADC_SEQA_IDX);
  // Clear all pending interrupts
  Chip_ADC_ClearFlags (LPC_ADC, Chip_ADC_GetFlags(LPC_ADC));
  // Enable sequencer
  Chip_ADC_EnableSequencer (LPC_ADC, ADC_SEQA_IDX);
  Chip_ADC_StartSequencer (LPC_ADC, ADC_SEQA_IDX);
}

/*======================================================================*/
/* USART                                                                */
/*======================================================================*/
#if DX_PROTOCOL_VER == 2
static const uint32_t baud[9] = { 38400, 57600, 115200, 1000000, 2000000, 3000000, 4000000, 4500000, 10500000 };
#endif
//-----------------------------
// タスクスイッチ励起
//-----------------------------
void us_csw (void) {
  taskYIELD ();
}

//-----------------------------
// 初期化
//-----------------------------
#if DX_PROTOCOL_VER == 1
static uint32_t us_init (uint32_t baudrate) {
//  usart0_dma_csw = us_csw;
  return usart0_dma_init (baudrate, txd, sizeof(txd), rxd, sizeof(rxd));
}
#elif DX_PROTOCOL_VER == 2
static uint32_t us_init (uint8_t ind) {
  if(ind > 4) ind = 4;
//  usart0_dma_csw = us_csw;
  return usart0_dma_init (baud[ind], txd, sizeof(txd), rxd, sizeof(rxd));
}
#endif

//-----------------------------
// ボーレート更新
//-----------------------------
#if DX_PROTOCOL_VER == 1
static uint32_t us_setbaudrate (uint32_t baudrate) {
  return usart0_dma_setbaudrate (baudrate);
}
#elif DX_PROTOCOL_VER == 2
static uint32_t us_setbaudrate (uint8_t ind) {
  if(ind > 4) ind = 4;
  return usart0_dma_setbaudrate (baud[ind]);
}
#endif

//-----------------------------
// 終了
//-----------------------------
static void us_deinit (void) { }

/*======================================================================*/
/* DX Communication task                                                */
/*======================================================================*/
//-----------------------------
// アイテムのレンジチェック(個別)
//-----------------------------
#if DX_PROTOCOL_VER == 1
static bool DX_CallbackValidPrevMapRange (const void *pmem, uint8_t u_st, uint8_t u_len) {
#elif DX_PROTOCOL_VER == 2
static bool DX_CallbackValidPrevMapRange (const void *pmem, uint16_t u_st, uint16_t u_len) {
#endif
  bool result = true;
  const TE097CtrlTable *pNewControlTable = pmem;

  for (int i = 0; i < 12; i++) {
    if (IS_SAFE_VAR(pNewControlTable, pNewControlTable->PinCfg[i], u_st, u_len)) {
      if (pNewControlTable->PinCfg[i] > 20) {
        result = false;
        break;
      } else for (int j = 0; j < 12; j++) {
        if ((i != j) && (pNewControlTable->PinCfg[i] >= tPinPWM0) && (pNewControlTable->PinCfg[i] <= tPinMPC3) && (pNewControlTable->PinCfg[i] == pmemmap->PinCfg[j])) {
          result = false;
          break;
        }
      }
    }
  }
  if (IS_SAFE_VAR(pNewControlTable, pNewControlTable->PWMFrequency, u_st, u_len)) {
    if (pNewControlTable->PWMFrequency > 50000) result = false;
  }
  if (IS_SAFE_VAR(pNewControlTable, pNewControlTable->OUT, u_st, u_len)) {
    if (pNewControlTable->OUT > 0xfff) result = false;
  }
  if (IS_SAFE_VAR(pNewControlTable, pNewControlTable->OUT, u_st, u_len)) {
    if (pNewControlTable->OUT > 0xfff) result = false;
  }
#if DX_PROTOCOL_VER == 2
  if (IS_SAFE_VAR(pNewControlTable, pNewControlTable->Baudrate, u_st, u_len)) {
    if (pNewControlTable->Baudrate > 4) result = false;
  }
#endif

  return result;
}

//-----------------------------
// アイテムのレンジチェック(一括)
//-----------------------------
bool DX_CallbackValidNextMapRange (const void *pmem) {
  const TE097CtrlTable *pNewControlTable = pmem;
  bool pincfg = false;

  for (int i = 0; i < 12; i++) {
    if (pNewControlTable->PinCfg[i] > 20) {
      pincfg = true;
      break;
    } else for (int j = 0; j < 12; j++) {
      if ((i != j) && (pNewControlTable->PinCfg[i] >= tPinPWM0) && (pNewControlTable->PinCfg[i] <= tPinMPC3) && (pNewControlTable->PinCfg[i] == pmemmap->PinCfg[j])) {
        pincfg = true;
        break;
      }
    }
  }

  return (
#if DX_PROTOCOL_VER == 2
    (pNewControlTable->Baudrate > 4)
    ||
#endif
    (pNewControlTable->OUT > 0xfff)
    || (pNewControlTable->OUT > 0xfff)
    || pincfg
  );
}

//-----------------------------
// NVM初期化時
//-----------------------------
void DX_CallbackInitNVMValue (void *pCtrl, const void *pNVM) {
  TE097CtrlTable *pControlTable = pCtrl;
  const TE097CtrlTable *pNVMdata = pNVM;

  update_PWMFreqency (0);

  pControlTable->PWMFrequency = pNVMdata->PWMFrequency;
  for (int i = 0; i < 6; i++) pControlTable->PWMDuty[i] = pNVMdata->PWMDuty[i];
  for (int i = 0; i < 12; i++) {
    pControlTable->PinCfg[i] = pNVMdata->PinCfg[i];
    switch (pControlTable->PinCfg[i]) {
      case tPinPWM0:    // 6
      case tPinPWM1:    // 7
      case tPinPWM2:    // 8
      case tPinPWM3:    // 9
      case tPinPWM4:    // 10
      case tPinPWM5:    // 11
        pin_config (i, pControlTable->PinCfg[i], pControlTable->PWMFrequency, pControlTable->PWMDuty[pControlTable->PinCfg[i] - tPinPWM0]);
        break;
      default:
        pin_config (i, pControlTable->PinCfg[i], pControlTable->PWMFrequency, 0);
        break;
    }
  }
}

//-----------------------------
// 読込パケット時
//-----------------------------
void DX_CallbackReadAction (void *pCtrl) {
  TE097CtrlTable *pControlTable = pCtrl;
  uint16_t w = 0;

  for (int i = 0; i < 12; i++) {
    pControlTable->ADV[i] = ADC_DR_RESULT (Chip_ADC_GetDataReg (LPC_ADC, i));
    w |= (Chip_GPIO_GetPinState (LPC_GPIO_PORT, 0, PinInfo[i].No) ? (1 << i) : 0);
  }
//  pControlTable->Capture[0] = pwd[0];
uint32_t r = LPC_SCT->CAP[4].U >> 1;
  pControlTable->Capture[0] = r;
  pControlTable->Capture[1] = r>>16;
  pControlTable->Capture[2] = pwd[2];
  pControlTable->Capture[3] = pwd[3];
  pControlTable->IN = w;
}

//-----------------------------
// 書込パケット時
//-----------------------------
void DX_CallbackWriteAction (const void *pNewCtrl, void *pCtrl, bool force) {
  const TE097CtrlTable *pNewControlTable = pNewCtrl;
  TE097CtrlTable *pControlTable = pCtrl;
  bool update_sct = false;

  // LED On/Off
  if (pNewControlTable->LED != pControlTable->LED || force) {
    SetLED (pNewControlTable->LED);
  }
  // Pin Config
  for (int i = 0; i < 12; i++) {
    if (pNewControlTable->PinCfg[i] != pControlTable->PinCfg[i] || force) {
      switch (pNewControlTable->PinCfg[i]) {
        case tPinPWM0:    // 6
        case tPinPWM1:    // 7
        case tPinPWM2:    // 8
        case tPinPWM3:    // 9
        case tPinPWM4:    // 10
        case tPinPWM5:    // 11
          pin_config (i, pNewControlTable->PinCfg[i], pNewControlTable->PWMFrequency, pNewControlTable->PWMDuty[pNewControlTable->PinCfg[i] - tPinPWM0]);
          break;
        default:
          pin_config (i, pNewControlTable->PinCfg[i], pNewControlTable->PWMFrequency, 0);
          break;
      }
    }
  }
  // PWM Frequency
  if (pNewControlTable->PWMFrequency != pControlTable->PWMFrequency || force) {
    update_PWMFreqency (pNewControlTable->PWMFrequency);
    update_sct = true;
  }
  // PWM Duty
  for (int i = 0; i < 6; i++) {
    if (pNewControlTable->PWMDuty[i] != pControlTable->PWMDuty[i] || force || update_sct) {
      update_PWMDuty (i, pNewControlTable->PWMDuty[i]);
    }
  }
  // OUT
  if (pNewControlTable->OUT != pControlTable->OUT || force) {
    pControlTable->OUT &= 0xfff;
    for (int i = 0; i < 12; i++) Chip_GPIO_SetPinState(LPC_GPIO_PORT, 0, PinInfo[i].No, pNewControlTable->OUT & (1 << i) ? true : false);
  }
}

/*======================================================================*/
/* FreeRTOS Stack over flow hook                                        */
/*======================================================================*/
void vApplicationStackOverflowHook (TaskHandle_t pxTask, char *pcTaskName) {
  taskDISABLE_INTERRUPTS ();
  while (1) {
    for (int i = 0; i < 1000; i++) {
      SetLED(true);
      Delay_us (1000 - i);
      SetLED (false);
      Delay_us (i);
      ToggleLED ();
    }
    Delay_ms (50);
    Chip_WWDT_Feed (LPC_WWDT);
  }
}

/*======================================================================*/
/* TASKs                                                                */
/*======================================================================*/
#ifdef _USE_MYTASK
void MyTask (void *pvParameters) {
  portTickType t;
  static int16_t p[6] = {4600,6800,9000,6800,6800,6800};
  static int16_t d[6] = { 15 ,-15 ,15, -15, 15, -15 };
  uint32_t tog = 0;


  vTaskDelay (1000);
#if DX_PROTOCOL_VER == 1
  DX_ReadMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
#elif DX_PROTOCOL_VER == 2
  DX2_ReadMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
#endif

  t = xTaskGetTickCount ();

  while (1) {
    for (int i = 0; i < 6; i++) {
      p[i] += d[i];
      if (p[i] > 11000) { p[i] = 11000; d[i] *= -1; }
      else if (p[i] < 4000) { p[i] = 4000; d[i] *= -1; }
    }
#if DX_PROTOCOL_VER == 1
    DX_WriteMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
#elif DX_PROTOCOL_VER == 2
    DX2_WriteMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
#endif

    if (tog++ % 10 == 0) ToggleLED ();

    vTaskDelayUntil (&t, 10);
  }
/*
  while (1) {
    if (p[0] != 32767) p[0] = 32767;
    else p[0] = 4096;
#if DX_PROTOCOL_VER == 1
    DX_WriteMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
#elif DX_PROTOCOL_VER == 2
    DX2_WriteMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
#endif
    vTaskDelayUntil (&t, 1000);
  }
*/
}
#endif

void TaskDXComm (void *pvParameters) {
#if DX_PROTOCOL_VER == 1
  DX_ClientCommTask ();
#elif DX_PROTOCOL_VER == 2
  DX2_ClientCommTask ();
#endif
}

/*======================================================================*/
/* main                                                                 */
/*======================================================================*/
void wdt_feed (void) { Chip_WWDT_Feed(LPC_WWDT); }
void sys_reset (void) { NVIC_SystemReset(); }

#if DX_PROTOCOL_VER == 1
static TDXClient_ConfParam ConfParam = {
  { (void *)&DEFAULT_DATA, sizeof(TE097CtrlTable), wdt_feed, sys_reset, 0x7F00, 256 },
  { us_init, us_deinit, us_setbaudrate, usart0_dma_tx_buff, usart0_dma_rx_buff, usart0_dma_tx_purge, usart0_dma_rx_purge, usart0_dma_putsbd, usart0_dma_getc },
  { DX_CallbackValidPrevMapRange, DX_CallbackValidNextMapRange, DX_CallbackInitNVMValue, DX_CallbackReadAction, DX_CallbackWriteAction }
};
#elif DX_PROTOCOL_VER == 2
static TDX2Client_ConfParam ConfParam = {
  { (void *)&DEFAULT_DATA, sizeof(TE097CtrlTable), true, NULL, 0, NULL, wdt_feed, sys_reset, 0x7F00, 256},
  { us_init, us_deinit, us_setbaudrate, usart0_dma_tx_buff, usart0_dma_rx_buff, usart0_dma_tx_purge, usart0_dma_rx_purge, usart0_dma_putsbd, usart0_dma_getc },
  { DX_CallbackValidPrevMapRange, DX_CallbackValidNextMapRange, DX_CallbackInitNVMValue, DX_CallbackReadAction, DX_CallbackWriteAction },
};
#endif

void main (void) {
  Chip_SYSCTL_SetBODLevels (SYSCTL_BODRSTLVL_1, SYSCTL_BODINTVAL_LVL1);
  Chip_SYSCTL_EnableBODReset ();

  PIO_Configure (pins, PIO_LISTSIZE (pins));
  init_wdt ();
  init_adc ();
	NVIC_EnableIRQ (SCT_IRQn);

  // DXクライアント機能設定
#if DX_PROTOCOL_VER == 1
  if (DX_Config (&ConfParam)) {
    pmemmap = DX_GetPCtrlTalbe ();
#elif DX_PROTOCOL_VER == 2
  if (DX2_Config (&ConfParam)) {
    pmemmap = DX2_GetPCtrlTalbe ();
#endif
    xTaskCreate (TaskDXComm, "", 210, NULL, 1, NULL);
#ifdef _USE_MYTASK
    xTaskCreate (MyTask, "", 210, NULL, 1, NULL);
#endif
    vTaskStartScheduler ();
  }
  while (1) {
    Chip_WWDT_Feed (LPC_WWDT);
    ToggleLED ();
    Delay_ms (200);
  }
}
